# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
# GTK UI for ramfile.
# Copyright 2005 by George A. Nelson.  All rights reserved.

import os
import sys
import re
import time

from ramfile_main import __author__, __email__, __url__, __copyright__, __version__
from ramfile_main import us, usdir, Go, Phase, Matching, font, geometry, url_opener, pasteOdd
from ramfile_main import help_file, help_geometry, help_tab_chars
from ramfile_main import CardDoc

import pygtk
if not hasattr(sys, 'frozen'):
    pygtk.require('2.0')
import gtk
if gtk.gtk_version[0] == 2 and gtk.gtk_version[1] < 4:
    sys.exit(_("""%s: gtk version 2.4 required""") % us)
import gobject, pango

def main():
    return gtk.main()

def gtkeat():
    '''Brute-force way to get our window updated.
    
    Note that self.window.process_updates(True) and
    gtk.gdk.window_process_all_updates() don't draw anything at all.
    '''
    while gtk.events_pending():
        gtk.main_iteration(False)   # should I check return value for quit?

def point_in_rect(x, y, r):
    rxy = gtk.gdk.Rectangle(x, y, 1, 1)
    r2 = r.intersect(rxy)
    return r2.width and r2.height

class RamfileWind(gtk.Window):
    '''GTK UI for Ramfile.
    
    Provides a menu bar, find/re popup, find edit field, notebook for
    documents, a text view for each document, and a status bar.  Much
    too much effort goes into managing the ordinary edit menu enabling.
    '''
    hand_cursor = gtk.gdk.Cursor(gtk.gdk.HAND2)
    regular_cursor = gtk.gdk.Cursor(gtk.gdk.XTERM)
    untitled_n = _("""Untitled""")
    readonly_n = _(""" - RO""")
    top_level_windows = []
    
    def quit_ok():
        for w in RamfileWind.top_level_windows:
            w.present()
            if not w.close_ok():
                return False
        return True
    quit_ok = staticmethod(quit_ok)
    
    def __init__(self, position=None):
        gtk.Window.__init__(self)
        
        self.loop = (None, 0)
        self.link_hovering = False
        self.matching = Matching()
        
        global font
        self.font = font
        self.font_desc = pango.FontDescription(font)
        #print self.font, self.font_desc, self.font_desc.get_family(), self.font_desc.to_string()
        
        #w = self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        #w.self = self
        RamfileWind.top_level_windows.append(self)
        self.connect('destroy', self.on_destroy)
        self.connect('delete_event', self.on_delete_event)
        self.set_border_width(2)
        self.set_default_size(400, 320)
        if geometry:
            self.parse_geometry(geometry)
        
        self.outer_box = box1 = gtk.VBox(False, 4)
        self.add(box1)
        box1.show()
        
        self.statusbar = sb = gtk.Statusbar()
        box1.pack_end(sb, False, False, 0)
        sb.show()
        sb.console_context = sb.get_context_id('console_message')
        
        self.make_ui_mgr(box1)
        
        box2 = gtk.HBox()
        box1.pack_start(box2, False, False, 0)
        box2.show()
        
        self.find_combo = fc = gtk.combo_box_new_text()
        for s in Matching.names:
            fc.append_text(s)
        fc.set_active(self.matching.item_n())
        box2.pack_start(fc, False, False, 4)
        fc.connect('changed', self.on_find_combo_changed)
        fc.show()
        
        frame = gtk.Frame()
        frame.show()
        box2.pack_start(frame)
        
        self.find_edit = fe = gtk.Entry(0)
        fe.set_editable(True)
        frame.add(fe)
        fe.sigids = [
            fe.connect('key-press-event', self.on_entry_keypress),
            fe.connect('focus-in-event', self.on_focusIn_find),
            fe.connect('focus-out-event', self.on_focusOut_find),
            fe.connect('changed', self.on_changed_find) ]
        fe.show()
        fe.orig_base_color = fe.get_style().base[gtk.STATE_NORMAL]
        fe.frame = frame
        fe.modify_font(self.font_desc)
        
        self.notebook = gtk.Notebook()
        box1.pack_start(self.notebook)
        self.notebook.set_scrollable(True)
        self.notebook.popup_enable()
        self.notebook.connect('switch-page', self.on_notebook_switch_page)
        self.notebook.show()
        
        self.set_main_title(-1)
        if position:
            self.set_position(position)             # this mostly doesn't work, no matter where I try it
        if geometry:
            self.resize(*self.get_default_size())   # why is this needed?
        self.show()
    
    def make_ui_mgr(self, box):
        self.uimgr = gtk.UIManager()
        self.add_accel_group(self.uimgr.get_accel_group())
        sb_context = self.statusbar.get_context_id('menu_tooltip')
        def install_menu_hints(uimgr, action, widget):
            if not isinstance(widget, gtk.MenuItem):
                return
            tip = action.get_property('tooltip')
            if tip:
                widget.connect('select', lambda item, tip=tip: self.statusbar.push(sb_context, tip))
                widget.connect('deselect', lambda item: self.statusbar.pop(sb_context))
        self.uimgr.connect('connect-proxy', install_menu_hints)
        
        selAll_n = _("""Select _All""")
        font_n = _("""_Font""")
        pasteOdd_n = _("""Paste _Odd Char""")
        
        ag = gtk.ActionGroup('main')
        ag.add_actions([
            ('File',    None, _("""_File""")),
            ('New',     gtk.STOCK_NEW, None, None, _("""New Card document"""), self.on_new),
            ('Open',    gtk.STOCK_OPEN, None, None, _("""Open a Card, MBox, or text file"""), self.on_open),
            ('Quit',    gtk.STOCK_QUIT, None, None, _("""Quit the Program; prompts to save"""), self.on_quit),
            
            ('Edit',    None, _("""_Edit""")),
            ('Matching', None, _("""_Matching""")),
            
            ('Help',    None, _("""_Help""")),
            ('HelpDoc', gtk.STOCK_HELP, _("""_Contents"""), None, _("""Help file"""), self.on_help),
            ('Pydoc', None, _("""Tech _Manual"""), None, _("""Technical data in Pydoc format"""), self.on_instruct),
            ('About', None, _("""_About"""), None, _("""About box"""), self.on_about),
        ])
        ag.add_radio_actions([
            ('MatchExact', None, _("""_Exact"""), None, _("""Upper and lower case are distinct in searches"""), Matching.exact),
            ('MatchCasemap', None, _("""_Case Map"""), None, _("""Upper and lower case are the same in searches"""), Matching.casemap),
            ('MatchAuto', None, _("""_Auto Case Map"""), None, _("""Searches map case if words are all lower case"""), Matching.auto),
        ], self.matching.mapping, self.on_matching)
        self.matching_radio_group = ag.get_action('MatchAuto').get_group()
        ag.add_toggle_actions([
            ('MatchRegex', None, _("""_Regular Expressions"""), None, _("""Pearl-Compatible Regular Expressions"""), self.on_regex),
        ])
        self.matching_re_toggle = ag.get_action('MatchRegex')
        self.uimgr.insert_action_group(ag, 0)
        
        self.doc_ag = ag = gtk.ActionGroup('doc')
        ag.add_actions([
            # File menu
            ('Close',   gtk.STOCK_CLOSE, None, None, _("""Close a document"""), self.on_close),
            ('Save',    gtk.STOCK_SAVE, None, None, _("""Save a document"""), self.on_save),
            ('SaveAs',  gtk.STOCK_SAVE_AS, None, None, _("""Save a document with a new name"""), self.on_saveAs),
            ('Revert',  gtk.STOCK_REVERT_TO_SAVED, None, None, _("""Revert to last saved version"""), self.on_revert),
            
            # Edit menu
            ('AddCard', gtk.STOCK_ADD, _("""Add Ca_rd"""), '<Control>r', _("""Add a new card after this card"""), self.on_addCard),
            ('DeleteCard', gtk.STOCK_REMOVE, _("""De_lete Card"""), '<Control>l', _("""Remove current card"""), self.on_delCard),
            ('RevertCard', gtk.STOCK_REFRESH, _("""R_evert Card"""), None, _("""Re-get current card"""), self.on_revertCard),
            ('GoHome',  gtk.STOCK_GOTO_TOP, _("""Go _Home"""), '<Control>h', _("""Display first card"""), self.on_goHome),
            ('Flip',  None, _("""F_ind/Card"""), '<Control>f', _("""Flip between Find and Card"""), self.on_flip_doc),
        ])
        self.uimgr.insert_action_group(ag, 2)
        ag.set_property('sensitive', False)
        
        self.none_edit_ag = ag = gtk.ActionGroup('none-edit')
        ag.add_actions([
            # Edit menu
            ('Cut',     gtk.STOCK_CUT, None, None, None, self.on_none),
            ('Copy',    gtk.STOCK_COPY, None, None, None, self.on_none),
            ('Paste',   gtk.STOCK_PASTE, None, None, None, self.on_none),
            ('Delete',  gtk.STOCK_DELETE, None, None, None, self.on_none),
            ('SelectAll', None, selAll_n, '<Control>a', None, self.on_none),
            ('Font',    None, font_n, None, None, self.on_none),
            ('PasteOdd', None, pasteOdd_n),
        ])
        self.uimgr.insert_action_group(ag, 3)
        ag.set_property('sensitive', False)
        
        self.doc_edit_ag = ag = gtk.ActionGroup('doc-edit')
        ag.add_actions([
            # Edit menu
            ('Cut',     gtk.STOCK_CUT, None, None, _("""Cut from Card"""), self.on_cut_doc),
            ('Copy',    gtk.STOCK_COPY, None, None, _("""Copy from Card"""), self.on_copy_doc),
            ('Paste',   gtk.STOCK_PASTE, None, None, _("""Paste to Card"""), self.on_paste_doc),
            ('Delete',  gtk.STOCK_DELETE, None, None, _("""Delete from Card"""), self.on_delete_doc),
            ('SelectAll', None, selAll_n, '<Control>a', _("""Select all text in Card"""), self.on_selectAll_doc),
            ('Font',    None, font_n, None, None, self.on_setFont_doc),
            ('PasteOdd', None, pasteOdd_n),
        ])
        #self.uimgr.insert_action_group(ag, 4)
        
        self.find_edit_ag = ag = gtk.ActionGroup('find-edit')
        ag.add_actions([
            # Edit menu
            ('Cut',     gtk.STOCK_CUT, None, None, _("""Cut from Find"""), self.on_cut_find),
            ('Copy',    gtk.STOCK_COPY, None, None, _("""Copy from Find"""), self.on_copy_find),
            ('Paste',   gtk.STOCK_PASTE, None, None, _("""Paste to Find"""), self.on_paste_find),
            ('Delete',  gtk.STOCK_DELETE, None, None, _("""Delete from Find"""), self.on_delete_find),
            ('SelectAll', None, selAll_n, '<Control>a', _("""Select all text in Find"""), self.on_selectAll_find),
            ('Font',    None, font_n, None, None, self.on_setFont_find),
            ('PasteOdd', None, pasteOdd_n),
        ])
        #self.uimgr.insert_action_group(ag, 4)
        
        merge_id = self.uimgr.add_ui_from_string('''
            <ui>
                <menubar>
                    <menu action="File">
                        <menuitem action="New" />
                        <menuitem action="Open" />
                        <menuitem action="Close" />
                        <separator/>
                        <menuitem action="Save" />
                        <menuitem action="SaveAs" />
                        <menuitem action="Revert" />
                        <separator/>
                        <placeholder name="FileMenuAdditions" />
                        <separator/>
                        <menuitem action="Quit" />
                    </menu>
                    <menu action="Edit">
                        <menuitem action="Cut"/>
                        <menuitem action="Copy"/>
                        <menuitem action="Paste"/>
                        <menuitem action="Delete"/>
                        <separator/>
                        <menuitem action="SelectAll"/>
                        <menuitem action="Font"/>
                        <menu action="PasteOdd">
                            <placeholder name="PasteOddAdditions"/>
                        </menu>
                        <menu action="Matching">
                            <menuitem action="MatchExact"/>
                            <menuitem action="MatchCasemap"/>
                            <menuitem action="MatchAuto"/>
                            <separator/>
                            <menuitem action="MatchRegex"/>
                        </menu>
                        <placeholder name="EditMenuSelectionAdditions" />
                        <separator/>
                        <menuitem action="AddCard"/>
                        <menuitem action="DeleteCard"/>
                        <menuitem action="RevertCard"/>
                        <menuitem action="GoHome"/>
                        <menuitem action="Flip"/>
                        <separator/>
                        <placeholder name="EditMenuAdditions" />
                    </menu>
                    <menu action="Help">
                        <menuitem action="HelpDoc"/>
                        <menuitem action="Pydoc"/>
                        <menuitem action="About"/>
                    </menu>
                </menubar>
            </ui>''')
        
        pasteOdd_merge_id = self.uimgr.new_merge_id()
        n_addn = 1
        for addn in pasteOdd:
            if isinstance(addn, str):
                addn = (addn, addn)
            if ( isinstance(addn, tuple) and len(addn) >= 2
                and isinstance(addn[0], str) and isinstance(addn[1], str) ):
                name = 'pasteOdd'+str(n_addn)
                def make_action(ag, callback):
                    action = gtk.Action(name, addn[0], '', None)
                    action.connect('activate', callback, addn[1])
                    if len(addn) is 3 and isinstance(addn[2], str):
                        action.set_property('tooltip', addn[2])
                    ag.add_action(action)
                make_action(self.none_edit_ag, self.on_none)
                make_action(self.doc_edit_ag, lambda b, c: self.doc_view_insert(c))
                make_action(self.find_edit_ag, lambda b, c: self.find_edit_insert(c))
                self.uimgr.add_ui( pasteOdd_merge_id,
                    'ui/menubar/Edit/PasteOdd/PasteOddAdditions',
                    name, name, gtk.UI_MANAGER_MENUITEM, False)        
            else:
                print _("""%s: bad pasteOdd setting "%s":\n"""
                        """    item #%d "%s" is not a string or a tuple of (name, string)""") % (
                        us, pasteOdd, n_addn, repr(addn) )
            n_addn += 1
        
        mbar = self.uimgr.get_widget('/menubar')
        mbar.show()
        box.pack_start(mbar, False, False, 0)
    
    def make_doc_view(self, doc, **kw):
        '''Display doc in a new notebook page after current page.
        '''
        RamfileNotebookPage(self, doc, **kw)
    
    def remove_doc_view(self, page_num):
        sw = self.notebook.get_nth_page(page_num)
        if sw:
            self.notebook.remove_page(page_num)
            sw.remove_doc_view(page_num)
            if not self.cur_page():
                self.doc_ag.set_property('sensitive', False)
    
    def set_main_title(self, page_num=None):
        if page_num is None:
            page_num = self.notebook.get_current_page()
        ttl = us
        sw = self.notebook.get_nth_page(page_num)
        if sw:
            path = os.path.basename(sw.page_doc.path)
            if not path:
                path = self.untitled_n
            if sw.page_doc.dirty or sw.page_buffer.get_modified():
                path += '*'
            if sw.page_doc.readonly:
                path += self.readonly_n
            ttl = '%s - %s' % (path, ttl)
        if self.get_title() != ttl:
            self.set_title(ttl)
    
    def set_tab_title(self, page_num=None):
        if page_num is None:
            page_num = self.notebook.get_current_page()
        sw = self.notebook.get_nth_page(page_num)
        if sw:
            tab = os.path.basename(sw.page_doc.path)
            if not tab:
                tab = self.untitled_n
            if sw.page_doc.dirty or sw.page_buffer.get_modified():
                tab += '*'
            if sw.page_doc.readonly:
                tab += self.readonly_n
            if sw.tab_label.get_text() != tab:
                sw.tab_label.set_text(tab)
                self.notebook.set_menu_label_text(sw, tab)
    
    def cur_page(self):
        '''Return current page's RamfileNotebookPage(), a ScrollWindow().
        
        RamfileNotebookPage() has page_doc, page_view, and page_buffer.
        Return None if no current page.
        '''
        page_num = self.notebook.get_current_page()
        if page_num >= 0:
            return self.notebook.get_nth_page(page_num)
    
    def set_notebook_page(self, i):
        was_focus = self.find_edit.is_focus()
        self.notebook.set_current_page(i)
        if was_focus:
            self.find_edit.grab_focus()
    
    def save_card(self):
        sw = self.cur_page()
        doc = sw.page_doc
        buf = sw.page_buffer
        if buf.get_modified():
            text = buf.get_text(*buf.get_bounds())
            if doc.get(doc.curCard) != text:
                doc.put(text, doc.curCard)
            buf.set_modified(False)
    
    def get_card(self, ci):
        sw = self.cur_page()
        doc = sw.page_doc
        buf = sw.page_buffer
        if ci != doc.curCard:
            buf.not_modified = not buf.get_modified()
            self.save_card()
            card = doc.get(ci)
            buf.set_text(card)
            buf.set_modified(False)
            buf.not_modified = False
            doc.curCard = ci
            self.doc_ag.get_action('GoHome').set_property('sensitive', ci != 0)
            #self.links_mark(card, buf)
            sw.page_view.scroll_to_iter(buf.get_start_iter(), 0.0)
            return card
    
    def reget_card(self):
        # When you really need to refresh the card, say after revert card or
        # delete card.  Does not save card.
        sw = self.cur_page()
        doc = sw.page_doc
        buf = sw.page_buffer
        card = doc.get(doc.curCard)
        buf.not_modified = not buf.get_modified()
        buf.set_text(card)
        buf.set_modified(False)
        buf.not_modified = False
        return card
    
    def links_mark_extend(self, buf, start_mark, end_mark):
        start_iter = buf.get_iter_at_mark(start_mark)
        start_iter.backward_chars(start_iter.get_line_offset())
        end_iter = buf.get_iter_at_mark(end_mark)
        if not end_iter.ends_line():
            end_iter.forward_to_line_end()
        text = buf.get_text(start_iter, end_iter)
        index = 0
        for line in xrange(start_iter.get_line()):
            index += buf.get_iter_at_line(line).get_bytes_in_line()
        #print 'links_mark_extend %d "%s"' % (index, text)
        self.links_mark(text, buf, index, start_iter, end_iter)
    
    def links_mark(self, text, buf, index=0, prev_iter=None, stop_iter=None):
        '''Hyperlinks
        
        In-card links start with "•".  URL links have "://" or "word@" or
        "word.", or are bounded by "<>" and contain at least one of ".@:". 
        Links are bounded by whitespace or "<>".  Links are all on one line. 
        "•link". "bar@foo.com", "http://www.foo.com", and
        "<ftp://ftp.foo.com> are all links.  This is all somewhat heuristic;
        some valid URLs will still need <> to be recognized, and some non-
        URLs will still slip through the TLD checks.
        '''
        # <http://data.iana.org/TLD/tlds-alpha-by-domain.txt>
        # Version 2007082201, Last Updated Thu Aug 23 08:07:00 2007 UTC
        ctlds = (
            ' AC AD AE AF AG AI AL AM AN AO AQ AR AS AT AU AW AX AZ BA BB'
            ' BD BE BF BG BH BI BJ BM BN BO BR BS BT BV BW BY BZ CA CC CD'
            ' CF CG CH CI CK CL CM CN CO CR CU CV CX CY CZ DE DJ DK DM DO'
            ' DZ EC EE EG ER ES ET EU FI FJ FK FM FO FR GA GB GD GE GF GG'
            ' GH GI GL GM GN GP GQ GR GS GT GU GW GY HK HM HN HR HT HU ID'
            ' IE IL IM IN IO IQ IR IS IT JE JM JO JP KE KG KH KI KM KN KR'
            ' KW KY KZ LA LB LC LI LK LR LS LT LU LV LY MA MC MD MG MH MK'
            ' ML MM MN MO MP MQ MR MS MT MU MV MW MX MY MZ NA NC NE NF NG'
            ' NI NL NO NP NR NU NZ OM PA PE PF PG PH PK PL PM PN PR PS PT'
            ' PW PY QA RE RO RU RW SA SB SC SD SE SG SH SI SJ SK SL SM SN'
            ' SO SR ST SU SV SY SZ TC TD TF TG TH TJ TK TL TM TN TO TP TR'
            ' TT TV TW TZ UA UG UK UM US UY UZ VA VC VE VG VI VN VU WF WS'
            ' YE YT YU ZA ZM ZW ' )
        gtlds = ( ' AERO ARPA ASIA BIZ CAT COM COOP CYM EDU GEO GOV INFO INT'
            ' JOBS MIL MOBI MUSEUM NAME NET ORG POST PRO TEL TRAVEL ' )
        lnk = ( '[-\w%]+((?:[.][-\w%]+)+)'      # .dom in a group
                '(?:\:\d+)?' )                  # allows :port
        pq = '(?:/[-\w%?&+=~#/.]+)?'            # /path ?query
        link_re = re.compile( '('
            '(?:•\S+)'                          # sloppy in-card link
            '|(?:[-\w.~%+:]+@'+lnk+')'          # email@name.dom
            '|(?:(?:[-\w%]+:/*)?'+lnk+pq+')'    # name.dom or scheme://name.dom
            '|(?:<\S+[.@:]\S+>)'                # <link>
            ')' )
        dom_re = re.compile('^[a-zA-Z]{2,4}$')  # for tld in tld.ctld
        line = 0
        line_index = 0
        if not prev_iter:
            prev_iter = buf.get_start_iter()
        if not stop_iter:
            stop_iter = buf.get_end_iter()
        for m in link_re.finditer(text):
            start, end = m.span(1)
            link = m.group(1)
            tld = m.group(2)
            if not tld:
                tld = m.group(3)
            if tld:
                tldl = tld.split('.')
                dom = ' %s ' % tldl[-1].upper()
                if ( len(tldl) > 2
                    and dom in ctlds
                    and dom_re.match(tldl[-2]) ):
                    pass
                elif not dom in gtlds:
                    continue
            if link[0] == '<' and link[-1] == '>':
                start += 1
                end -= 1
            start += index
            end += index
            while True: # Gah.  No API to make iter at byte, only byte in line.
                bytes = buf.get_iter_at_line(line).get_bytes_in_line()
                if line_index + bytes > start:
                    break
                line += 1
                line_index += bytes
            #print start, end, line, line_index, '"'+link+'"'
            start_iter = buf.get_iter_at_line_index(line, start-line_index)
            end_iter = buf.get_iter_at_line_index(line, end-line_index)
            buf.remove_tag_by_name('link-tag', prev_iter, start_iter)
            buf.apply_tag_by_name('link-tag', start_iter, end_iter)
            prev_iter = end_iter
        buf.remove_tag_by_name('link-tag', prev_iter, stop_iter)
        sw = self.cur_page()
        wx, wy, mod = sw.page_view.window.get_pointer()
        x, y = sw.page_view.window_to_buffer_coords(gtk.TEXT_WINDOW_WIDGET, wx, wy)
        self.links_set_cursor(sw.page_view, x, y)
    
    def links_set_cursor(self, text_view, x, y):
        hovering = False
        test_hover = False
        if getattr(text_view, 'get_iter_at_position', None):
            iter1 = text_view.get_iter_at_position(x, y)[0]
        else:
            iter1 = text_view.get_iter_at_location(x, y)
            # In gtk < 2.8 the only way to track uses "locations between
            # characters", so tracking is offset half a char to the front
            # (usually left); this is especially apparent when the preceding
            # character is a Tab.
            test_hover = True
        for tag in iter1.get_tags():
            if tag.get_property('name') == 'link-tag':
                if test_hover:
                    iter2 = iter1.copy()    # test left boundary
                    if iter2.backward_char() and not iter2.has_tag(tag):
                        r = text_view.get_iter_location(iter2)
                        if point_in_rect(x, y, r):
                            continue
                hovering = True
                break
        if test_hover and not hovering and iter1.backward_char():
            for tag in iter1.get_tags():    # test right boundaryc 
                if tag.get_property('name') == 'link-tag':
                    r = text_view.get_iter_location(iter1)
                    if point_in_rect(x, y, r):
                        hovering = True
                        break
        text_view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(
            hovering and self.hand_cursor or self.regular_cursor )
        text_view.window.get_pointer()      # causes link to get MOTION_NOTIFY?
        self.link_hovering = hovering
    
    def show_looped(self, is_looped, not_found):
        #print 'looped'
        self.find_edit.modify_base( gtk.STATE_NORMAL,
            is_looped and
                ( not_found and gtk.gdk.Color(0xFFFF, 0xC000, 0xC000)
                    or gtk.gdk.Color(0xC000, 0xFFFF, 0xC000) )
            or self.find_edit.orig_base_color )
    
    def do_find(self, findStr, go, matching):
        self.save_card
        sw = self.cur_page()
        if findStr != self.loop[0]:
            self.loop = (findStr, sw.page_doc)
            sw.page_doc.loopCard = sw.page_doc.curCard
            for i in xrange(self.notebook.get_n_pages()):
                sw = self.notebook.get_nth_page(i)
                buf = sw.page_buffer
                buf.remove_tag_by_name( 'found-text',
                    buf.get_start_iter(), buf.get_end_iter() )
        
        #   1) search current doc from "next" to "end"
        #   2) search any other docs from "start" to "end"
        #   3) search current doc from "start" thru "cur"
        #
        #   search in current direction, so "start" and "end" may be swapped.
        
        st = time.time()
        np = self.notebook.get_n_pages()
        cp = self.notebook.get_current_page()
        ci = self.do_find_1(cp, findStr, go, matching, Phase.next)
        if ci is not None:
            self.do_found(ci, findStr, matching)
            return
        pageList = range(cp+1,np)+range(0,cp)
        if go == Go.up:
            pl1 = range(0,cp) ; pl1.reverse()
            pl2 = range(cp+1,np) ; pl2.reverse()
            pageList = pl1 + pl2
        for i in pageList:
            #self.set_notebook_page(i)
            #gtkeat()
            ci = self.do_find_1(i, findStr, go, matching, Phase.all)
            if ci is not None:
                self.set_notebook_page(i)
                self.do_found(ci, findStr, matching)
                return
        #self.set_notebook_page(cp)
        #gtkeat()
        ci = self.do_find_1(cp, findStr, go, matching, Phase.rest)
        if ci is not None:
            self.do_found(ci, findStr, matching)
            return
        et = time.time() - st
        if et > .2:
            print '%s: not found time %6.2f' % (us, et)
    
    def do_find_1(self, pageNum, findStr, go, matching, phase):
        sw = self.notebook.get_nth_page(pageNum)    #self.cur_page()
        lbl = self.notebook.get_tab_label(sw)
        lbl.set_finding(True)
        doc = sw.page_doc
        ci, looped = doc.find(findStr, go, matching, phase)
        lbl.set_finding(False)
        if looped and self.loop[1] == doc:
            self.show_looped(True, ci is None and phase is Phase.rest)
        return ci
    
    def do_found(self, ci, findStr, matching):
        sw = self.cur_page()
        doc = sw.page_doc
        buf = sw.page_buffer
        card = self.get_card(ci)
        if card:
            # Color the found words.
            wl = doc.find_positions(card, findStr, matching)
            wl.sort()
            line = 0
            line_index = 0
            for pos, length in wl:
                while True: # Gah.  No API to make iter at byte, only byte in line.
                    bytes = buf.get_iter_at_line(line).get_bytes_in_line()
                    if line_index + bytes > pos:
                        break
                    line += 1
                    line_index += bytes
                buf.apply_tag_by_name( 'found-text',
                    buf.get_iter_at_line_index(line, pos-line_index),
                    buf.get_iter_at_line_index(line, pos+length-line_index) )
        return True
    
    def close_save_test(self):
        self.save_card()
        sw = self.cur_page()
        if sw.page_doc.dirty:
            path = sw.page_doc.path
            if not path:
                path = self.untitled_n
            else:
                path = os.path.basename(path)
            dlg = gtk.MessageDialog( self,
                gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
                _("""Save changes to "%s" before closing?""") % path )
            dlg.add_button(_("""Cancel"""), gtk.RESPONSE_CANCEL)
            res = dlg.run()
            dlg.destroy()
            if res == int(gtk.RESPONSE_CANCEL):
                return False
            elif res == int(gtk.RESPONSE_YES):
                self.on_save(None)
            elif res == int(gtk.RESPONSE_NO):
                sw.page_doc.dirty = False
        return True
    
    def close_ok(self):
        for i in xrange(self.notebook.get_n_pages()):
            self.set_notebook_page(i)
            if not self.close_save_test():
                return False
        return True
    
    def filter_text(self, filter_info):
        path, uri, disp, mime = filter_info
        if ( mime.startswith('text/')
            or mime == 'application/postscript'
            or mime == 'application/x-desktop'
            or mime == 'application/x-perl'
            or mime == 'application/x-python'
            or mime == 'application/x-php'
            or mime == 'application/x-shellscript'
            or mime == 'application/x-troff' ):
            return True
    
    def find_edit_insert(self, text):
        fe = self.find_edit
        fe.delete_selection()
        pos = fe.get_position()
        fe.insert_text(text, pos)
        fe.set_position(pos+len(text))
    
    def doc_view_insert(self, text):
        sw = self.cur_page()
        buf = sw.page_buffer
        editable = sw.page_view.get_editable()
        buf.delete_selection(True, editable)
        buf.insert_interactive_at_cursor(text, editable)
    
    def console_status(self, s):
        print s
        self.console_status_clear()
        sb = self.statusbar
        msg_id = sb.push(sb.console_context, s)
        gobject.timeout_add( 3000,
            lambda self=self, msg_id=msg_id: self.console_status_clear(msg_id) )
    
    def console_status_clear(self, msg_id=None):
        sb = self.statusbar
        if msg_id:
            sb.remove(sb.console_context, msg_id)
        else:
            sb.pop(sb.console_context)
    
    def drag_indicate(self, w, i, r, is_on=False):
        if not r:
            #print 'drag_indicate no rect', w, i, is_on
            return
        nbw = w.notebook.window
        if not is_on:
            nbw.invalidate_rect(r, True)
        else:
            gc = nbw.new_gc()
            gc.set_rgb_fg_color(gtk.gdk.Color(0x0000, 0xFFFF, 0x0000))
            #gc.set_rgb_bg_color(gtk.gdk.Color(0x0000, 0xFFFF, 0x0000))
            nbw.draw_rectangle(gc, True, r.x, r.y, r.width, r.height)
    
    def open_link(self, link, go):
        if link.startswith('•'):            # note: '•' is more than one byte long
            self.find_edit.set_text(link)
            self.find_edit.grab_focus()
            self.do_find( link, go, Matching() )
        else:
            if ':' in link:
                pass                        # assume URL scheme set
            elif '@' in link:
                link = 'mailto:%s' % link   # "mailto:" doesn't use "//"
            elif link.startswith('ftp.'):
                link = 'ftp://%s' % link    # otherwise tends to open web page
            try:
                #os.startfile(link) -- only MSWindows, only files, not URLs
                status = None
                if os.path.exists(url_opener):
                    status = os.spawnl( os.P_WAIT, url_opener,
                        os.path.basename(url_opener), link )
                else:
                    try:
                        import webbrowser
                    except:
                        pass
                    else:
                        status = webbrowser.open_new(link)
                if status:
                    self.console_status( _("""%s: link "%s" failed, exit status=%d""")
                        % (us, link, status) )
            except:
                self.console_status( _("""%s: exception opening link "%s": %s""")
                    % (us, link, str(sys.exc_info()[1])) )
    
    def ui_check_readonly(self, sw):
        wr = sw and not sw.page_doc.readonly
        self.doc_ag.get_action('Save').set_property('sensitive', wr)
        self.doc_ag.get_action('Revert').set_property('sensitive', wr)
        self.doc_ag.get_action('AddCard').set_property('sensitive', wr)
        self.doc_ag.get_action('DeleteCard').set_property('sensitive', wr)
        self.doc_ag.get_action('RevertCard').set_property('sensitive', wr)
        self.doc_edit_ag.get_action('Cut').set_property('sensitive', wr)
        self.doc_edit_ag.get_action('Paste').set_property('sensitive', wr)
        self.doc_edit_ag.get_action('Delete').set_property('sensitive', wr)
        self.doc_edit_ag.get_action('PasteOdd').set_property('sensitive', wr)
    
    def set_font(self, who):
        dlg = gtk.FontSelectionDialog(_("""Set %s Font""") % who)
        if self.font:
            dlg.set_font_name(self.font)
        res = dlg.run()
        fn = dlg.get_font_name()
        dlg.destroy()
        if res == int(gtk.RESPONSE_OK):
            self.console_status(fn)
            self.font = fn
            self.font_desc = pango.FontDescription(fn)
            #self.find_edit.modify_font(self.font_desc)
            return True
    
    #   Callbacks:
    
    def on_find_combo_changed(self, b):
        m = b.get_active()
        self.matching.set(m)
        for r in self.matching_radio_group:
            r.set_active(self.matching.mapping == r.get_property('value'))
        self.matching_re_toggle.set_active(self.matching.use_regex)
    
    def on_entry_keypress(self, widget, event, data=None):
        # Need the event for Shift state, so can't use Activate signal.
        if event.type == gtk.gdk.KEY_PRESS:
            if event.keyval == gtk.keysyms.Return:
                fe = self.find_edit
                fe.select_region(0,-1)
                self.show_looped(False, None)
                self.do_find( fe.get_text(),
                    (event.state & gtk.gdk.SHIFT_MASK) and Go.up or Go.down,
                    self.matching )
                return True
    
    def on_changed_find(self, editable, data=None):
        fe = self.find_edit
        if fe.get_style().base[gtk.STATE_NORMAL] != fe.orig_base_color:
            self.show_looped(False, None)
            self.loop = (None, 0)
    
    def on_notebook_switch_page(self, notebook, gtkpage, page_num, data=None):
        #sw = self.cur_page() -- doesn't work on first page added;
        #   Notebook.get_current_page() returns -1 instead of 0
        sw = notebook.get_nth_page(page_num)
        modified = sw and sw.page_buffer.get_modified()
        self.doc_ag.get_action('RevertCard').set_property('sensitive', modified)
        self.ui_check_readonly(sw)
        self.set_main_title(page_num)
    
    def on_destroy(self, wid):
        #print "on_destroy"
        RamfileWind.top_level_windows.remove(self)
        #self.window.self = None
        if not len(RamfileWind.top_level_windows):
            gtk.main_quit()
        else:
            gtkeat()    # eww.  Prevent signal delivery to destroyed window.
    
    def on_delete_event(self, a1, a2):
        #print "on_delete_event"
        if not self.close_ok():
            return True
        #def discon(o):
        #    for sid in o.sigids:
        #        o.disconnect(sid)
        #discon(self.find_edit)
        #self.find_edit.disconnect(self.find_edit.sigids[2])
        #sw = self.cur_page()
        #if sw:
        #    discon(sw.page_view)
    
    def on_focusIn_find(self, widget, event, data=None):
        self.uimgr.insert_action_group(self.find_edit_ag, 4)
        self.uimgr.remove_action_group(self.none_edit_ag)
        widget.frame.set_shadow_type(gtk.SHADOW_IN)
    
    def on_focusOut_find(self, widget, event, data=None):
        self.uimgr.insert_action_group(self.none_edit_ag, 3)
        self.uimgr.remove_action_group(self.find_edit_ag)
        widget.frame.set_shadow_type(gtk.SHADOW_NONE)
    
    def on_focusIn_doc(self, widget, event, data=None):
        self.uimgr.insert_action_group(self.doc_edit_ag, 4)
        self.uimgr.remove_action_group(self.none_edit_ag)
        return
    
    def on_focusOut_doc(self, widget, event, data=None):
        self.uimgr.insert_action_group(self.none_edit_ag, 3)
        self.uimgr.remove_action_group(self.doc_edit_ag)
    
    def on_bufDirty(self, buf, data=None):
        buf = self.cur_page().page_buffer
        if buf.not_modified:
            return
        self.doc_ag.get_action('RevertCard').set_property( 'sensitive',
            buf.get_modified() )
        self.set_main_title()
        self.set_tab_title()
    
    def on_bufDelete(self, buf, start_iter, end_iter, data=None):
        buf.create_mark('bufDeleteStart', start_iter, True)
        buf.create_mark('bufDeleteEnd', end_iter, False)
    
    def on_bufInsert(self, buf, start_iter, text, length, data=None):
        buf.create_mark('bufInsertStart', start_iter, True)
        buf.create_mark('bufInsertEnd', start_iter, False)
    
    def on_bufChanged(self, buf, data=None):
        #print 'bufChanged'
        delStart = buf.get_mark('bufDeleteStart')
        delEnd = buf.get_mark('bufDeleteEnd')
        insStart = buf.get_mark('bufInsertStart')
        insEnd = buf.get_mark('bufInsertEnd')
        if delStart is not None:
            if delEnd is None:
                print '%s: bufChanged bufDeleteStart but no bufDeleteEnd' % us
            else:
                start = buf.get_iter_at_mark(delStart).get_offset()
                end = buf.get_iter_at_mark(delEnd).get_offset()
                if start != end:
                    print '%s: bufChanged bufDeleteStart=%d != bufDeleteEnd=%d' % (us, start, end)
                else:
                    self.links_mark_extend(buf, delStart, delStart)
                buf.delete_mark(delEnd)
            buf.delete_mark(delStart)
        elif insStart is not None:
            if insEnd is None:
                print '%s: bufChanged bufInsertStart but no bufInsertEnd' % us
            else:
                start = buf.get_iter_at_mark(insStart).get_offset()
                end = buf.get_iter_at_mark(insEnd).get_offset()
                self.links_mark_extend(buf, insStart, insEnd)
                buf.delete_mark(insEnd)
            buf.delete_mark(insStart)
        else:
            print '%s: bufChanged not prepared' % us
    
    def on_link_event(self, texttag, widget, event, iter, data=None):
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 1 and self.link_hovering:
            self.console_status_clear()
            sw = self.cur_page()
            start_iter = iter.copy()
            #start_iter.backward_to_tag_toggle(texttag)
            while start_iter.get_offset() > sw.page_buffer.get_start_iter().get_offset():
                if not start_iter.has_tag(texttag):
                    start_iter.forward_char()
                    break
                start_iter.backward_char()
            end_iter = iter.copy()
            end_iter.forward_to_tag_toggle(texttag)
            go = (event.state & gtk.gdk.SHIFT_MASK) and Go.up or Go.down
            self.open_link(sw.page_buffer.get_text(start_iter, end_iter), go)
            return True
    
    def on_motion_doc(self, text_view, event):
        x, y = text_view.window_to_buffer_coords( gtk.TEXT_WINDOW_WIDGET,
            int(event.x), int(event.y) )
        self.links_set_cursor(text_view, x, y)
    
    def on_visibility_doc(self, text_view, event):
        wx, wy, mod = text_view.window.get_pointer()
        x, y = text_view.window_to_buffer_coords(gtk.TEXT_WINDOW_WIDGET, wx, wy)
        self.links_set_cursor(text_view, x, y)
    
    def on_tab_close_doc(self, b, sw):
        page_num = self.notebook.page_num(sw)
        self.notebook.set_current_page(page_num)
        if not self.close_save_test():
            return
        self.find_edit.grab_focus()
        self.remove_doc_view(page_num)
    
    def on_tab_dragging(self, event, sw, widget):
        new_target = None
        new_target_n = -1
        new_target_rect = None
        force_draw = False
        target_w = gtk.gdk.window_at_pointer()  # bleah
        if target_w:
            target_w = target_w[0].get_toplevel()
        for w in RamfileWind.top_level_windows:
            if w.window == target_w:
                ws = w #w.self
                wnb = ws.notebook
                rnb = wnb.allocation
                new_target = ws
                target_sw = ws.cur_page()
                if target_sw is None:
                    new_target_n = 0
                else:
                    rt = gtk.gdk.Rectangle( rnb.x, rnb.y,
                        0, target_sw.page_view.allocation.y-rnb.y )
                    new_target_rect = gtk.gdk.Rectangle(
                        -(2 + wnb.get_property('tab-hborder') + 1),
                        rt.y+1, 3, rt.height-2 )
                    #x, y = int(event.x), int(event.y)
                    x, y, mods = target_sw.window.get_pointer()
                    if point_in_rect(x, y, target_sw.page_view.allocation):
                        # in notebook page
                        new_target_n = wnb.page_num(sw)
                        new_target_rect.x += wnb.get_tab_label(wnb.get_nth_page(new_target_n)).allocation.x
                    elif point_in_rect(x, y, wnb.allocation):
                        # in tab area
                        # this won't (?) handle rtl locale
                        new_target_n = -1
                        first_vis = -1; last_vis = -1
                        for i in xrange(wnb.get_n_pages()):
                            swt = wnb.get_nth_page(i)
                            tlt = wnb.get_tab_label(swt)
                            if not tlt.get_child_visible():
                                #print 'cont', i
                                continue
                            tlr = tlt.allocation
                            if first_vis == -1:
                                first_vis = i
                                rt.width = tlr.x - rt.x
                                if point_in_rect(x, y, rt):
                                    if i > 0:
                                        new_target_n = 0
                                        new_target_rect.x = rnb.x
                                        new_target_rect.width = 6
                                        force_draw = True
                                        break
                                else:
                                    rt.x = tlr.x
                            last_vis = i
                            rt.width = tlr.x + tlr.width/2 - rt.x
                            if point_in_rect(x, y, rt):
                                new_target_n = i
                                new_target_rect.x += tlr.x
                                break
                            rt.x += rt.width
                        rt.width = tlr.x + tlr.width - rt.x             # right edge of last label
                        if last_vis+1 == wnb.get_n_pages():             # last page is visible,
                            rt.width = rnb.x + rnb.width - rt.x         #   so right edge of tabs area
                        if new_target_n < 0 and point_in_rect(x, y, rt):
                            new_target_n = last_vis+1
                            new_target_rect.x = tlr.x + tlr.width + 1
                        rt.x += rt.width                                # past last test
                        rt.width = rnb.x + rnb.width - rt.x             # right edge of tabs area
                        if new_target_n < 0 and point_in_rect(x, y, rt):
                            new_target_n = wnb.get_n_pages()
                            new_target_rect.x = rnb.x + rnb.width - 6
                            new_target_rect.width = 6
                            force_draw = True                           # overlaps tabs scroll btn
                    if new_target_n < 0:
                        new_target_n = wnb.get_n_pages()
                        swt = wnb.get_nth_page(new_target_n-1)
                        tlt = wnb.get_tab_label(swt)
                        if not tlt.get_child_visible():
                            new_target_rect.x = rnb.x + rnb.width - 6
                            new_target_rect.width = 6
                        else:
                            tlr = tlt.allocation
                            new_target_rect.x = tlr.x + tlr.width + 1
                if ( new_target != widget.drag_target
                    or new_target_n != widget.drag_target_n ):
                    if widget.drag_target:
                        self.drag_indicate( widget.drag_target,
                            widget.drag_target_n, widget.drag_target_rect )
                    self.drag_indicate( new_target,
                        new_target_n, new_target_rect, True )
                elif force_draw:
                    self.drag_indicate(new_target,
                        new_target_n, new_target_rect, True)
                widget.drag_target = new_target
                widget.drag_target_n = new_target_n
                widget.drag_target_rect = new_target_rect
                return
        if widget.drag_target:
            self.drag_indicate( widget.drag_target,
                widget.drag_target_n, widget.drag_target_rect )
        widget.drag_target = new_target
        widget.drag_target_n = new_target_n
        widget.drag_target_rect = new_target_rect
    
    def on_tab_dragged(self, event, sw, widget):
        #print 'drop at', widget.drag_target_n
        n = widget.drag_target_n
        t = widget.drag_target
        if t:
            self.drag_indicate(t, n, widget.drag_target_rect)
        nb = self.notebook
        if t == self:
            # same window
            if n >= 0:
                swp = nb.page_num(sw)
                if n > swp:
                    n -= 1
                np = nb.get_n_pages()
                if n >= np:
                    n = np-1
                nb.reorder_child(sw, n)
        else: # different or new window
            tl = nb.get_tab_label(sw)
            nb.set_tab_label(sw)
            tl.unparent()
            nb.remove(sw)
            sw.unparent()
            if not t:
                t = RamfileWind(gtk.WIN_POS_MOUSE)
                t.find_edit.grab_focus()
            sw.owner = t
            t.notebook.insert_page(sw, tl, n)
            if not self.cur_page():
                self.destroy()
        sw.owner.set_notebook_page(sw.owner.notebook.page_num(sw))
        widget.drag_target = None
        widget.drag_target_n = -1
        widget.drag_target_rect = None
    
    def on_none(self, b):
        pass
    
    #   Menu item callbacks:
    
    def on_new(self, b):
        doc = CardDoc.newdoc()
        if doc:
            self.make_doc_view(doc)
            self.cur_page().page_view.grab_focus()
    
    def on_open(self, b):
        chooser = gtk.FileChooserDialog( _("""Open"""), None,
            gtk.FILE_CHOOSER_ACTION_OPEN,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK) )
        chooser.set_default_response(gtk.RESPONSE_OK)
        chooser.set_select_multiple(True)
        sw = self.cur_page()
        if sw and sw.page_doc.path:
            path, name = os.path.split(sw.page_doc.path)
            chooser.set_current_folder(path)
        
        filt = gtk.FileFilter()
        filt.set_name(_("""Most Text Files"""))
        filt.add_custom(gtk.FILE_FILTER_MIME_TYPE, self.filter_text)
        chooser.add_filter(filt)

        filt = gtk.FileFilter()
        filt.set_name(_("""All Files"""))
        filt.add_pattern('*')
        chooser.add_filter(filt)
        
        paths = []
        if chooser.run() == int(gtk.RESPONSE_OK):
            paths = chooser.get_filenames()
        chooser.destroy()
        
        for path in paths:
            if os.path.isfile(path):
                doc, ld = CardDoc.opendoc(path)
                if ld == CardDoc.ld_enc_err:
                    dlg = gtk.MessageDialog( self,
                        gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK_CANCEL,
                        _("""Encoding error: file "%s" is not %s.  Some data will be lost.""")
                            % (os.path.basename(path), doc.encoding) )
                    res = dlg.run()
                    dlg.destroy()
                    if res == int(gtk.RESPONSE_CANCEL):
                        doc = None
                elif ld == CardDoc.ld_no_file:
                    dlg = gtk.MessageDialog( self,
                        gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_CANCEL,
                        _("""Can't open file "%s".""") % path )
                    res = dlg.run()
                    dlg.destroy()
                if doc:
                    self.make_doc_view(doc)
                    self.find_edit.grab_focus()
    
    def on_close(self, b):
        if not self.close_save_test():
            return
        page_num = self.notebook.get_current_page()
        if page_num >= 0:
            self.find_edit.grab_focus()
            self.remove_doc_view(page_num)
    
    def on_save(self, b):
        self.save_card()
        sw = self.cur_page()
        if sw.page_doc.path:
            sw.page_doc.save()
            self.set_main_title()
            self.set_tab_title()
        else:
            self.on_saveAs(b)
    
    def on_saveAs(self, b):
        sw = self.cur_page()
        doc = sw.page_doc
        
        chooser = gtk.FileChooserDialog( _("""Save As"""), None,
            gtk.FILE_CHOOSER_ACTION_SAVE,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE, gtk.RESPONSE_OK) )
        chooser.set_default_response(gtk.RESPONSE_OK)
        if getattr(chooser, 'set_do_overwrite_confirmation', None):
            chooser.set_do_overwrite_confirmation(True)
        if doc.path:
            path, name = os.path.split(doc.path)
            chooser.set_current_folder(path)
            chooser.set_current_name(name)
        
        filt = gtk.FileFilter()
        filt.set_name(_("""Most Text Files"""))
        filt.add_custom(gtk.FILE_FILTER_MIME_TYPE, self.filter_text)
        chooser.add_filter(filt)

        filt = gtk.FileFilter()
        filt.set_name(_("""All Files"""))
        filt.add_pattern('*')
        chooser.add_filter(filt)
        
        path = None
        if chooser.run() == int(gtk.RESPONSE_OK):
            path = chooser.get_filename()
        chooser.destroy()
        
        if path:
            self.save_card()
            doc.save_as(path)
            ttl = os.path.basename(doc.path)
            self.ui_check_readonly(sw)
            self.set_main_title()
            self.set_tab_title()
    
    def on_revert(self, b):
        self.save_card()
        sw = self.cur_page()
        doc = sw.page_doc
        buf = sw.page_buffer
        save_curCard = doc.curCard
        doc.revert()
        doc.set_curCard(save_curCard)
        self.reget_card()                       # might not be the original card
        self.find_edit.grab_focus()
        self.set_main_title()
        self.set_tab_title()
    
    def on_quit(self, b):
        if self.quit_ok():
            gtk.main_quit()
    
    def on_cut_find(self, b):
        self.find_edit.cut_clipboard()
    
    def on_cut_doc(self, b):
        sw = self.cur_page()
        sw.page_buffer.cut_clipboard(gtk.clipboard_get(), sw.page_view.get_editable())
        sw.page_view.scroll_mark_onscreen(sw.page_buffer.get_insert())
    
    def on_copy_find(self, b):
        self.find_edit.copy_clipboard()
    
    def on_copy_doc(self, b):
        sw = self.cur_page()
        sw.page_buffer.copy_clipboard(gtk.clipboard_get())
    
    def on_paste_find(self, b):
        self.find_edit.paste_clipboard()
    
    def on_paste_doc(self, b):
        sw = self.cur_page()
        sw.page_buffer.paste_clipboard(gtk.clipboard_get(), None, sw.page_view.get_editable())
        sw.page_view.scroll_mark_onscreen(sw.page_buffer.get_insert())
    
    def on_delete_find(self, b):
        self.find_edit.delete_selection()
    
    def on_delete_doc(self, b):
        sw = self.cur_page()
        sw.page_buffer.delete_selection(True, sw.page_view.get_editable())
        sw.page_view.scroll_mark_onscreen(sw.page_buffer.get_insert())
    
    def on_selectAll_find(self, b):
        self.find_edit.select_region(0, -1)
    
    def on_selectAll_doc(self, b):
        sw = self.cur_page()
        buf = sw.page_buffer
        #buf.select_range(buf.get_start_iter(), buf.get_end_iter())
        #   select_range() doesn't work unless applied twice.  Selects start to
        #   cursor pos the first time
        buf.begin_user_action()
        buf.move_mark_by_name('insert', buf.get_start_iter())
        buf.move_mark_by_name('selection_bound', buf.get_end_iter())
        buf.end_user_action()
    
    def on_setFont_find(self, b):
        if self.set_font(_("""find box""")):
            self.find_edit.modify_font(self.font_desc)
    
    def on_setFont_doc(self, b):
        sw = self.cur_page()
        who = os.path.basename(sw.page_doc.path)
        if not who:
            who = self.untitled_n
        if self.set_font(who):
            sw.page_view.modify_font(self.font_desc)
            #--redo the tabs
    
    def on_flip_doc(self, b):
        if self.find_edit.is_focus():
            self.cur_page().page_view.grab_focus()
        else:
            self.find_edit.grab_focus()
    
    def on_matching(self, newRA, curRA):
        self.matching.mapping = m = newRA.get_current_value()
        self.find_combo.set_active(self.matching.item_n())
    
    def on_regex(self, action):
        self.matching.use_regex = action.get_active()
        self.find_combo.set_active(self.matching.item_n())
    
    def on_addCard(self, b):
        sw = self.cur_page()
        doc = sw.page_doc
        ci = doc.curCard+1
        doc.add(ci)
        self.get_card(ci)
        self.cur_page().page_view.grab_focus()
    
    def on_delCard(self, b):
        sw = self.cur_page()
        doc = sw.page_doc
        doc.delete(doc.curCard)
        self.reget_card()
        sw.page_view.scroll_to_iter(sw.page_buffer.get_start_iter(), 0.0)
        self.find_edit.grab_focus()
    
    def on_revertCard(self, b):
        self.reget_card()
    
    def on_goHome(self, b):
        self.get_card(0)
        self.find_edit.grab_focus()
    
    def on_help(self, b):
        #path = os.path.join(os.path.dirname(sys.argv[0]), help_file)
        path = os.path.join(usdir, help_file)
        print path
        if os.path.exists(path):
            doc, ld = CardDoc.opendoc(path)
            if ld == CardDoc.ld_enc_err:
                print _("""%s: Encoding error: file "%s" is not %s.""") % (
                    us, path, doc.encoding )
            elif ld == CardDoc.ld_no_file:
                print _("""%s: can't open "%s" """) % (us, path)
            elif doc:
                doc.readonly = True
                global geometry
                save_geometry = geometry
                geometry = help_geometry
                w = RamfileWind()
                geometry = save_geometry
                w.make_doc_view(doc, tab_chars=help_tab_chars)
                w.find_edit.grab_focus()
    
    def on_instruct(self, b):
        w = self.instructWind = gtk.Window(gtk.WINDOW_TOPLEVEL)
        w.set_border_width(2)
        
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        
        tv = gtk.TextView()
        tv.set_editable(False)
        tv.set_wrap_mode(gtk.WRAP_NONE)
        tv.set_left_margin(4)
        tv.set_right_margin(4)
        sw.add(tv)
        tv.show()
        
        buf = tv.get_buffer()
        import pydoc
        d = pydoc.TextDoc()
        d.bold = lambda text: chr(2) + text + chr(3)
        t = d.docmodule(sys.modules[__name__])
        if not t:
            t = '--broken--'
        t = t.replace(__name__, us)     # woot!
        monospaceFont_n = _("""Monospace""")
        plain = buf.create_tag(font=monospaceFont_n)
        bold = buf.create_tag(font=monospaceFont_n, weight=700)
        start = 0
        while start < len(t):
            end = t.find(chr(2), start)
            if end < 0:
                end = len(t)
            buf.insert_with_tags(buf.get_end_iter(), t[start:end], plain)
            end += 1
            start = t.find(chr(3), end)
            if start < 0:
                start = len(t)
            buf.insert_with_tags(buf.get_end_iter(), t[end:start], bold)
            start += 1
        buf.set_modified(False)
        
        w.add(sw)
        sw.show()
        
        r = tv.get_iter_location(buf.get_iter_at_offset(0))
        lw = 84*r.width + 2
        lh = 30*r.height + 2
        tv.set_size_request(lw, lh)
        w.set_title(_("""Ramfile Tech Manual"""))
        w.show()

    def on_about(self, b):
        def add_link_btn(l):
            def open_link(b, l):
                self.open_link(l, Go.down)
                return True
            lbl = gtk.Label()
            lbl.set_markup(_("""<span color='blue' underline='single'>%s</span>""") % l)
            btn = gtk.Button()
            btn.add(lbl)
            btn.set_relief(gtk.RELIEF_NONE)
            btn.set_focus_on_click(False)
            btn.connect('clicked', open_link, l)
            btn.set_alignment(0.0, 0.5)
            dlg.vbox.pack_start(btn, False, False, 0)
        
        dlg = gtk.Dialog(_("""About ramfile"""), self, gtk.DIALOG_MODAL)
        ok_btn = dlg.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        
        hbox = gtk.HBox()
        lbl = gtk.Label()
        lbl.set_markup( _("""<span size='large' weight='bold'>ramfile v.%s</span>""")
            % (__version__) )
        hbox.pack_start(lbl, False, False, 0)
        lbl = gtk.Label()
        lbl.set_markup( _("""<span size='large' weight='bold'>%s</span>""")
            % (__author__) )
        hbox.pack_end(lbl, False, False, 0)
        dlg.vbox.pack_start(hbox, False, False, 10)
        
        lbl = gtk.Label()
        lbl.set_markup(_("""%s""") % __copyright__)
        lbl.set_alignment(0.0, 0.5)
        dlg.vbox.pack_start(lbl, False, False, 6)
        
        add_link_btn(__url__)
        add_link_btn(__email__)
        
        ok_btn.grab_focus()
        dlg.vbox.show_all()
        res = dlg.run()
        dlg.destroy()

class NotebookTabLabel(gtk.HBox):
    '''Fancy "draggable" Notebook tab label with close button.
    '''
    drag_cursor = gtk.gdk.Cursor(gtk.gdk.FLEUR)
    
    def __init__(self, on_close, on_dragging, on_drag_end, owner_):
        gtk.HBox.__init__(self, False, 0)
        
        event_box = gtk.EventBox()
        event_box.set_visible_window(False)
        self.pack_start(event_box)
        event_box.show()
        
        label = self.label = gtk.Label()
        label.set_alignment(0.0, 0.5)
        event_box.add(label)
        label.orig_fg = label.get_style().fg[gtk.STATE_NORMAL]
        label.show()
        
        close_image = gtk.image_new_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
        image_w, image_h = gtk.icon_size_lookup(gtk.ICON_SIZE_MENU)
        
        close_btn = gtk.Button()
        close_btn.set_relief(gtk.RELIEF_NONE)
        close_btn.connect('clicked', on_close, owner_)
        close_btn.set_size_request(image_w+2, image_h+2)
        close_btn.add(close_image)
        self.pack_start(close_btn, False, False)
        close_btn.show_all()
        
        self.show()
        
        def on_enter(widget, event):
            if hasattr(owner_, 'owner') and hasattr(owner_.owner, 'statusbar'):
                statusbar = owner_.owner.statusbar
                sb_context = statusbar.get_context_id('notebooktablabel')
                statusbar.push(sb_context, owner_.page_doc.path)
        def on_leave(widget, event):
            if hasattr(owner_, 'owner') and hasattr(owner_.owner, 'statusbar'):
                statusbar = owner_.owner.statusbar
                sb_context = statusbar.get_context_id('notebooktablabel')
                statusbar.pop(sb_context)
        event_box.connect('enter-notify-event', on_enter)
        event_box.connect('leave-notify-event', on_leave)
        
        def on_tab_motion(widget, event):
            if widget.in_drag:
                on_dragging(event, owner_, widget)
            elif widget.drag_check_threshold(
                    widget.x_start, widget.y_start,
                    int(event.x_root), int(event.y_root) ):
                widget.in_drag = True
                widget.grab_add()
                gtk.gdk.pointer_grab( widget.window, False,
                    gtk.gdk.POINTER_MOTION_MASK | gtk.gdk.BUTTON_RELEASE_MASK,
                    None, self.drag_cursor, long(event.time) )
                if hasattr(owner_, 'owner') and hasattr(owner_.owner, 'statusbar'):
                    statusbar = owner_.owner.statusbar
                    sb_context = statusbar.get_context_id('notebooktablabel')
                    statusbar.pop(sb_context)   # no leave event during grab
        
        def on_tab_btn_press(widget, event):
            if event.button == 1:
                widget.x_start = int(event.x_root)
                widget.y_start = int(event.y_root)
                widget.in_drag = False
                widget.drag_target = None
                widget.motion_signal = widget.connect('motion_notify_event', on_tab_motion)
        
        def on_tab_btn_release(widget, event):
            if event.button != 1:
                return
            widget.disconnect(widget.motion_signal)
            widget.motion_signal = None
            if widget.in_drag:
                widget.in_drag = False
                gtk.gdk.pointer_ungrab(long(event.time))
                widget.grab_remove()
                on_drag_end(event, owner_, widget)
        
        self.connect('button_press_event', on_tab_btn_press)
        self.connect('button_release_event', on_tab_btn_release)
    
    def set_finding(self, is_finding):
        lbl = self.label
        lbl.modify_fg( gtk.STATE_NORMAL,
            is_finding and gtk.gdk.Color(0x0000, 0x8000, 0x0000) or lbl.orig_fg )
        txt = lbl.get_text()
        lbl.set_text(txt+' ')
        lbl.set_text(txt)
        gtkeat()

class RamfileNotebookPage(gtk.ScrolledWindow):
    '''Self-contained ScrolledWindow for use as Ramfile Notebook page.
    '''
    def __init__(self, owner_, doc, tab_chars=8):
        '''Display doc in a new Notebook() page after current page.  Owner
        is the RamfileWind() containing the Notebook().
        '''
        gtk.ScrolledWindow.__init__(self)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        self.owner = owner_
        
        tv = gtk.TextView()
        tv.set_editable(not doc.readonly)
        tv.set_wrap_mode(gtk.WRAP_WORD)
        tv.set_accepts_tab(False)
        tv.set_left_margin(4)
        tv.set_right_margin(4)
        tv.set_indent(-10)
        self.add(tv)
        tv.show()
        tv.sigids = [
            tv.connect('focus-in-event', lambda *a: self.owner.on_focusIn_doc(*a)),
            tv.connect('focus-out-event', lambda *a: self.owner.on_focusOut_doc(*a)),
            tv.connect('motion-notify-event', lambda *a: self.owner.on_motion_doc(*a)),
            tv.connect('visibility-notify-event', lambda *a: self.owner.on_visibility_doc(*a)) ]
        if owner_.font_desc:
            tv.modify_font(owner_.font_desc)
        
        buf = tv.get_buffer()
        #
        buf.set_text('n')
        buf.set_modified(False)
        buf.not_modified = False
        r = tv.get_iter_location(buf.get_iter_at_offset(0))
        tabwidth = tab_chars*r.width
        tabstops = pango.TabArray(128, True)
        for i in xrange(128):
            tabstops.set_tab(i, pango.TAB_LEFT, tabwidth*i)
        tv.set_tabs(tabstops)
        #
        buf.connect('modified-changed', lambda *a: self.owner.on_bufDirty(*a))
        buf.connect('delete-range', lambda *a: self.owner.on_bufDelete(*a))
        buf.connect('insert-text', lambda *a: self.owner.on_bufInsert(*a))
        buf.connect('changed', lambda *a: self.owner.on_bufChanged(*a))
        buf.insert_length = None
        
        buf.create_tag('found-text',
            background_gdk=gtk.gdk.Color(0xE000, 0xF000, 0xFFFF),
            background_set=True )
        
        link_tag = buf.create_tag('link-tag',
            foreground_gdk=gtk.gdk.Color(0, 0, 0xFFFF),
            underline=pango.UNDERLINE_SINGLE )
        link_tag.connect('event', lambda *a: self.owner.on_link_event(*a))
        
        self.page_doc = doc
        self.page_view = tv
        self.page_buffer = buf
        
        self.show()
        nb = self.owner.notebook
        page = nb.insert_page(self, position=nb.get_current_page()+1)
        self.owner.set_notebook_page(page)
        tl = NotebookTabLabel(
            lambda *a: self.owner.on_tab_close_doc(*a),
            lambda *a: self.owner.on_tab_dragging(*a),
            lambda *a: self.owner.on_tab_dragged(*a),
            self )
        tl.drag_target_n = None
        tl.drag_target_rect = None
        self.tab_label = tl.label
        nb.set_tab_label(self, tl)
        self.owner.set_tab_title()
        self.owner.doc_ag.set_property('sensitive', True)
        #self.owner.doc_ag.get_action('RevertCard').set_property('sensitive', False)
        self.owner.doc_ag.get_action('GoHome').set_property('sensitive', False)
        doc.curCard = 0
        self.owner.reget_card()
    
    def remove_doc_view(self, page_num):
        del self.page_doc
        del self.page_view
        del self.page_buffer
        del self.owner
